part of masamune.data;

/// Document that stores [TemporaryField] with Key
/// 
/// Users do n’t really [TemporaryField].
/// 
/// Normal, Can be treated as if the data were handled [Map<String, dynamic>].
/// 
/// ```
/// TemporaryDocument document = TemporaryDocument.fromMap(
///   {"A": "Data is A", "B": "Data is B"}
/// );
/// document["A"] // Data is A;
/// document["C"] = "Data is C";
/// ```
class TemporaryDocument extends Document<TemporaryField>
  implements IDataDocument<TemporaryField> {
  static Iterable<TemporaryField> _convertData(
    Map<String, dynamic> data
  ) {
    List<TemporaryField> list = ListPool.get();
    data?.forEach(
      ( key, value ) {
        if( isEmpty(key) || value == null ) return;
        list.add(
          TemporaryField(value)
        );
      }
    );
    return list;
  }
  /// Process to create a new instance.
  /// 
  /// Do not use from outside the class.
  /// 
  /// [path]: Destination path.
  /// [isTemporary]: True if the data is temporary.
  @override
  @protected
  T createCloneInstance<T extends IClonable>( String path, bool isTemporary ) => TemporaryDocument._(
  ) as T;
  /// Document that stores [TemporaryField] with Key
  /// 
  /// Users do n’t really [TemporaryField].
  /// 
  /// Normal, Can be treated as if the data were handled [Map<String, dynamic>].
  /// 
  /// ```
  /// TemporaryDocument document = TemporaryDocument.fromMap(
  ///   {"A": "Data is A", "B": "Data is B"}
  /// );
  /// document["A"] // Data is A;
  /// document["C"] = "Data is C";
  /// ```
  factory TemporaryDocument(
  ){
    return TemporaryDocument._( );
  }
  /// Document that stores [TemporaryField] with Key
  /// 
  /// Users do n’t really [TemporaryField].
  /// 
  /// Normal, Can be treated as if the data were handled [Map<String, dynamic>].
  /// 
  /// ```
  /// TemporaryDocument document = TemporaryDocument.fromMap(
  ///   {"A": "Data is A", "B": "Data is B"}
  /// );
  /// document["A"] // Data is A;
  /// document["C"] = "Data is C";
  /// ```
  /// 
  /// [jsonData]: Json string representing Map.
  factory TemporaryDocument.fromJson(
    String jsonData
  ){
    Map<String, dynamic> data = Json.decodeAsMap(jsonData);
    return TemporaryDocument.fromMap(data);
  }
  /// Document that stores [TemporaryField] with Key
  /// 
  /// Users do n’t really [TemporaryField].
  /// 
  /// Normal, Can be treated as if the data were handled [Map<String, dynamic>].
  /// 
  /// ```
  /// TemporaryDocument document = TemporaryDocument.fromMap(
  ///   {"A": "Data is A", "B": "Data is B"}
  /// );
  /// document["A"] // Data is A;
  /// document["C"] = "Data is C";
  /// ```
  /// 
  /// [mapData]: Map data to be stored.
  factory TemporaryDocument.fromMap(
    Map<String, dynamic> mapData
  ){
    return TemporaryDocument._( children:_convertData( mapData ) );
  }
  TemporaryDocument._(
    {
      Iterable<TemporaryField> children
    }
  ) : super(
    null,
    children: children,
    isTemporary: true
  );
  

  /// Get the protocol of the path.
  @override
  String get protocol => Protocol.tmp;
  
  /// True if the object is temporary data.
  @override
  bool get isTemporary => true;


  /// Get the data.
  /// 
  /// Do not use from external class.
  /// 
  /// When using from an external class, use getInt or getString.
  /// 
  /// [key]: Key for retrieving data.
  @protected
  dynamic operator []( String key ){
    if( !this.data.containsKey( key ) ) return null;
    return this.data[key]?.data;
  }
  /// Set the data.
  /// 
  /// [key]: Key for storing data [value]: Data to store.
  void operator []=( String key, dynamic value ){
    if( value == null ) { return; }
    if( this.isLock ) { Log.error( "Data modification is prohibited." ); return; }
    if( value is TemporaryField ){
      this.set( [value] );
    } else {
      if( PathFilter.setTemporary( Paths.child(this.path, key), value) ) return;
      if( this.data.containsKey( key ) && this.data[key] is IUnit<dynamic> ) {
         this.data[key].data = value;
      } else {
        Path path = this.rawPath.clone( );
        path.path = Paths.child( this.rawPath.path, key );
        this.set( [TemporaryField( value )] );
      }
    }
  }

  /// Gets value as Bool.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  bool getBool(String key, [bool defaultValue = false]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    TemporaryField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getBool( defaultValue );
  }
  /// Gets value as Double.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  double getDouble(String key, [double defaultValue = 0]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    TemporaryField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getDouble( defaultValue );
  }
  /// Gets value as Int.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  int getInt(String key, [int defaultValue = 0]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    TemporaryField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getInt( defaultValue );
  }
  /// Gets value as List<T>.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  List<T> getList<T extends Object>(
    String key,
    [
      List<T> defaultValue = const []
    ]
  ) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    TemporaryField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getList<T>( defaultValue );
  }
  /// Gets value as Map<K,V>.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  Map<K, V> getMap<K extends Object, V extends Object>(
    String key,
    [
      Map<K, V> defaultValue = const {}
    ]
  ) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    TemporaryField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getMap<K,V>( defaultValue );
  }
  /// Gets value as String.
  /// 
  /// [key]: Key for retrieving data.
  /// [defaultValue]: Initial value if there is no value.
  @override
  String getString(String key, [String defaultValue = Const.empty]) {
    if( isEmpty(key) || !this.data.containsKey(key) ) return defaultValue;
    TemporaryField data = this.data[key];
    if( data == null ) return defaultValue;
    return this.data[key].getString( defaultValue );
  }


  /// Get the UID of the document.
  /// 
  /// If there is no value in the field, id will be output.
  String get uid => this.getString(Const.uid, this.id);
  /// Get time.
  /// 
  /// UpdatedTime is output if the field has no value.
  int get time => this.getInt(Const.time, this.updatedTime);


  /// Save the data.
  /// 
  /// Run if you have a remote or need to save data.
  Future<T> save<T extends IDataDocument>( ) => Future<T>.delayed(Duration.zero);
  /// Delete the data.
  /// 
  /// Used when deleting data when there is a remote or when data needs to be saved.
  Future delete( ){
    if( this.isDisposed ) return Future.delayed(Duration.zero);
    this.dispose();
    return Future.delayed(Duration.zero);
  }
}